@@ -39,6 +39,7 @@ Implements several helper classes to handle the state of a session
 and the communication with the module processes
 """
 
+#import asyncio.exceptions
 import base64
 import errno
 import gzip
@@ -52,6 +53,7 @@ import time
 import uuid
 from ipaddress import ip_address
 
+import concurrent.futures
 import ldap
 import pycurl
 import six
@@ -116,15 +118,19 @@ class ModuleConnection(object):
     def __init__(self):
         self._client = tornado.httpclient.AsyncHTTPClient()
 
-    @tornado.gen.coroutine
-    def connect(self, connect_retries=0):
+    def abort(self):
+        # tornado cannot cancel a running request: https://github.com/tornadoweb/tornado/issues/157
+        # We have to do it manually and ignore the exceptions which happen there
+        # this will lead to a "AttributeError: 'NoneType' object has no attribute 'socket_action'" in the main loop, which we can safely ignore
+        self._client.close()
+        self._client = tornado.httpclient.AsyncHTTPClient()
+
+    async def connect(self, connect_retries=0):
         pass
 
-    @tornado.gen.coroutine
     def request(self, method, uri, headers=None, body=None):
         pass
 
-    @tornado.gen.coroutine
     def do_request(self, method, uri, headers=None, body=None, unix_socket=None):
         request = tornado.httpclient.HTTPRequest(
             self.get_uri(uri),
@@ -138,8 +144,38 @@ class ModuleConnection(object):
             prepare_curl_callback=(lambda curl: curl.setopt(pycurl.UNIX_SOCKET_PATH, unix_socket)) if unix_socket else None,
         )
 
+        return self._wrap_future(self._client.fetch(request, raise_error=True))
+
+    def _wrap_future(self, request_future):
+        result_future = tornado.concurrent.Future()
+
+        def propagate_result(future):
+            if future.cancelled():
+                if not result_future.cancelled():
+                    result_future.cancel()
+            elif future.exception():
+                def reraise():
+                    raise future.exception()
+                try:
+                    response = self._handle_errors(reraise)
+                except Exception as exc:
+                    result_future.set_exception(exc)
+                else:
+                    result_future.set_result(response)
+            else:
+                result_future.set_result(future.result())
+
+        def cancel_downstream(future):
+            if future.cancelled() and not request_future.cancelled():
+                request_future.cancel()
+
+        request_future.add_done_callback(propagate_result)
+        result_future.add_done_callback(cancel_downstream)
+        return result_future
+
+    def _handle_errors(self, function):
         try:
-            response = yield self._client.fetch(request, raise_error=True)
+            response = function()
         except tornado.curl_httpclient.CurlError as exc:
             CORE.warn('Reaching module failed: %s' % (exc,))
             raise CouldNotConnect(exc)
@@ -150,8 +186,14 @@ class ModuleConnection(object):
         except ValueError as exc:  # HTTP GET request with body
             CORE.warn('Reaching module failed: %s' % (exc,))
             raise BadRequest(str(exc))
+        except concurrent.futures.CancelledError as exc:
+            CORE.warn('Aborted module process request: %s' % (exc,))
+            raise CouldNotConnect(exc)
+        #except asyncio.exceptions.CancelledError as exc:
+        #    CORE.warn('Aborted module process request: %s' % (exc,))
+        #    raise CouldNotConnect(exc)
 
-        raise tornado.gen.Return(response)
+        return response
 
     def get_uri(self, uri):
         return uri
@@ -193,10 +235,9 @@ class ModuleProcess(ModuleConnection):
     def set_exit_callback(self, callback):
         self.__process.set_exit_callback(callback)
 
-    @tornado.gen.coroutine
-    def connect(self, connect_retries=0):
+    async def connect(self, connect_retries=0):
         if os.path.exists(self.socket) and stat.S_ISSOCK(os.stat(self.socket).st_mode):
-            raise tornado.gen.Return(True)
+            return True
         elif connect_retries > 200:
             raise CouldNotConnect('timeout exceeded')
         elif self.__process and self.__process.proc.poll() is not None:
@@ -209,10 +250,9 @@ class ModuleProcess(ModuleConnection):
             if connect_retries and not connect_retries % 50:
                 CORE.info('No connection to module process yet')
             connect_retries += 1
-            yield tornado.gen.sleep(0.05)
-            yield self.connect(connect_retries)
+            await tornado.gen.sleep(0.05)
+            await self.connect(connect_retries)
 
-    @tornado.gen.coroutine
     def request(self, method, uri, headers=None, body=None):
         # watch the module's activity and kill it after X seconds inactivity
         self.reset_inactivity_timer()
@@ -220,12 +260,12 @@ class ModuleProcess(ModuleConnection):
         self._active_requests.add(request_id)
 
         try:
-            response = yield self.do_request(method, uri, headers, body, self.socket)
+            response = self.do_request(method, uri, headers, body, self.socket)
         finally:
             self._active_requests.remove(request_id)
             self.reset_inactivity_timer()
 
-        raise tornado.gen.Return(response)
+        return response
 
     def get_uri(self, uri):
         if uri.startswith('https://'):
@@ -239,12 +279,11 @@ class ModuleProcess(ModuleConnection):
         if self.__process:
             tornado.ioloop.IOLoop.current().add_callback(self.stop_process)
 
-    @tornado.gen.coroutine
-    def stop_process(self):
+    async def stop_process(self):
         proc = self.__process.proc
         if proc.poll() is None:
             proc.terminate()
-        yield tornado.gen.sleep(3.0)
+        await tornado.gen.sleep(3.0)
         if proc.poll() is None:
             proc.kill()
         CORE.info('ModuleProcess: child stopped')
@@ -314,14 +353,11 @@ class ModuleProxy(ModuleConnection):
         self.proxy_address = proxy_address
         self.unix_socket = None
 
-    @tornado.gen.coroutine
-    def connect(self, connect_retries=0):
-        raise tornado.gen.Return(not self.unix_socket or os.path.exists(self.unix_socket))
+    async def connect(self, connect_retries=0):
+        return not self.unix_socket or os.path.exists(self.unix_socket)
 
-    @tornado.gen.coroutine
     def request(self, method, uri, headers=None, body=None):
-        response = yield self.do_request(method, uri, headers, body, self.unix_socket)
-        raise tornado.gen.Return(response)
+        return self.do_request(method, uri, headers, body, self.unix_socket)
 
     def get_uri(self, uri):
         request = urlsplit(uri)
@@ -359,9 +395,8 @@ class Nothing(Resource):
 
     requires_authentication = False
 
-    @tornado.gen.coroutine
-    def prepare(self, *args, **kwargs):
-        yield super(Nothing, self).prepare(*args, **kwargs)
+    async def prepare(self, *args, **kwargs):
+        await super(Nothing, self).prepare(*args, **kwargs)
         raise NotFound()
 
 
@@ -423,7 +458,7 @@ class Auth(Resource):
 
     requires_authentication = False
 
-    def parse_authorization(self):
+    async def parse_authorization(self):
         return  # do not call super method: prevent basic auth
 
     @sanitize(
@@ -432,8 +467,7 @@ class Auth(Resource):
         auth_type=StringSanitizer(allow_none=True),
         new_password=StringSanitizer(required=False, allow_none=True, minimum=1),
     )
-    @tornado.gen.coroutine
-    def post(self):
+    async def post(self):
         try:
             content_length = int(self.request.headers.get("Content-Length", 0))
         except ValueError:
@@ -454,7 +488,7 @@ class Auth(Resource):
         # important: must be called before the auth, to preserve session id in case of re-auth and that a user cannot choose his own session ID by providing a cookie
         sessionid = self.create_sessionid(True)
 
-        result = yield session.authenticate(self.request.body_arguments)
+        result = await session.authenticate(self.request.body_arguments)
 
         self.set_session(sessionid)
         self.set_status(result.status)
@@ -470,9 +504,8 @@ class Modules(Resource):
 
     requires_authentication = False
 
-    @tornado.gen.coroutine
-    def prepare(self):
-        yield super(Modules, self).prepare()
+    async def prepare(self):
+        await super(Modules, self).prepare()
         self.i18n = I18N_Manager()
         self.i18n['umc-core'] = I18N()
         self.i18n.set_locale(self.locale.code)
@@ -564,9 +597,8 @@ class Categories(Resource):
 
     requires_authentication = False
 
-    @tornado.gen.coroutine
-    def prepare(self):
-        yield super(Categories, self).prepare()
+    async def prepare(self):
+        await super(Categories, self).prepare()
         self.i18n = I18N_Manager()
         self.i18n['umc-core'] = I18N()
         self.i18n.set_locale(self.locale.code)
@@ -614,14 +646,39 @@ class Command(Resource):
 
     requires_authentication = False
 
+    async def prepare(self, *args, **kwargs):
+        await super(Command, self).prepare(*args, **kwargs)
+        self.future = None
+        self.process = None
+
     def forbidden_or_unauthenticated(self, message):
         # make sure that the UMC login dialog is shown if e.g. restarting the UMC-Server during active sessions
         if self.current_user.user.authenticated:
             return Forbidden(message)
         return Unauthorized(self._("For using this module a login is required."))
 
-    @tornado.gen.coroutine
-    def get(self, umcp_command, command):
+    def on_connection_close(self):
+        super(Command, self).on_connection_close()
+        CORE.warn('Connection was aborted by the client!')
+        self._remove_active_request()
+        if self.future is not None:
+            self.future.cancel()
+        if self.process is not None:
+            self.process.abort()
+
+    def on_finish(self):
+        super(Command, self).on_finish()
+        self._remove_active_request()
+
+    def _remove_active_request(self):
+        session = self.current_user
+        if session and session._active_requests:
+            try:
+                session._active_requests.remove(hash(self))
+            except KeyError:
+                pass
+
+    async def get(self, umcp_command, command):
         """
         Handles a COMMAND request. The request must contain a valid
         and known command that can be accessed by the current user. If
@@ -637,6 +694,7 @@ class Command(Resource):
         """
         session = self.current_user
         acls = session.acls
+        session._active_requests.add(hash(self))
 
         module_name = acls.get_module_providing(moduleManager, command)
         if not module_name:
@@ -661,15 +719,18 @@ class Command(Resource):
         locale = Locale(self.locale.code)
         if not locale.territory:  # TODO: replace by using the actual provided value
             locale.territory = {'de': 'DE', 'fr': 'FR', 'en': 'US'}.get(self.locale.code)
-        process = session.processes.get_process(module_name, str(locale), self.settings.get("no_daemonize_module_processes"))
-        request_id = uuid.uuid4()
-        session._active_requests.add(request_id)
+        process = self.process = session.processes.get_process(module_name, str(locale), self.settings.get("no_daemonize_module_processes"))
         CORE.info('Passing request to module %s' % (module_name,))
 
         try:
-            yield process.connect()
+            await process.connect()
             # send first command
-            response = yield process.request(self.request.method, self.request.full_url(), body=self.request.body or None, headers=headers)
+            self.future = process.request(self.request.method, self.request.full_url(), body=self.request.body or None, headers=headers)
+            response = await self.future
+        except concurrent.futures.CancelledError:
+            raise BadGateway('%s: %s: canceled' % (self._('Connection to module process failed'), module_name))
+        #except asyncio.exceptions.CancelledError:
+        #    raise BadGateway('%s: %s: canceled' % (self._('Connection to module process failed'), module_name))
         except CouldNotConnect as exc:
             # (happens during starting the service and subprocesses when the UNIX sockets aren't available yet)
             # also happens when module process gets killed during request
@@ -703,8 +764,6 @@ class Command(Resource):
                 self.set_header('Content-Length', str(len(response.body)))
                 self.write(response.body)
             self.finish()
-        finally:
-            session._active_requests.remove(request_id)
 
     def get_request_header(self, session, methodname, umcp_command):
         headers = dict(self.request.headers)
@@ -880,8 +939,7 @@ class Set(Resource):
         use specific pathes ("set/{password,locale,user/preferences}") instead
     """
 
-    @tornado.gen.coroutine
-    def post(self):
+    async def post(self):
         is_univention_lib = self.request.headers.get('User-Agent', '').startswith('UCS/')
         for key in self.request.body_arguments:
             cls = {'password': SetPassword, 'user': SetUserPreferences, 'locale': SetLocale}.get(key)
@@ -891,7 +949,7 @@ class Set(Resource):
                 p = cls(self.application, self.request)
                 p._ = self._
                 p.finish = self.finish
-                yield p.post()
+                await p.post()
                 return
             if key == 'password':
                 self.redirect('/univention/set/password', status=307)
@@ -933,8 +991,7 @@ class SetPassword(Resource):
         "password": StringSanitizer(required=True),
         "new_password": StringSanitizer(required=True),
     }))
-    @tornado.gen.coroutine
-    def post(self):
+    async def post(self):
         assert self.current_user.user.authenticated
         username = self.current_user.user.username
         password = self.request.body_arguments['password']['password']
@@ -949,7 +1006,7 @@ class SetPassword(Resource):
 
         CORE.info('Changing password of user %r' % (username,))
         try:
-            yield self.current_user.change_password(args)
+            await self.current_user.change_password(args)
         except PasswordChangeFailed as exc:
             raise UMC_Error(str(exc), 400, {'new_password': '%s' % (exc,)})  # 422
         else:
